
#ifndef _DAMLDB_MMALOC_ALLOCATOR_HH
#define _DAMLDB_MMALOC_ALLOCATOR_HH

#if defined (sun)
#  pragma ident "@(#)$Id: MMallocAllocator.hh,v 1.2 2002/04/09 02:04:46 neves Exp $"
#endif

#include "damldb/Portability.hh"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <string>
#include <stdexcept>

#if !defined(HAVE_MMALOC_H)
extern "C" {
  void* mmalloc_attach(int fd, void* baseaddr);
  void* mmalloc_detach(void* md);
  void* mmalloc(void* md, size_t size);
  void* mrealloc(void*md, void* ptr, size_t size);
  void* mvalloc(void* md, size_t size);
  void* mfree(void*md, void* ptr);
  void* mmalloc_errno(void* md);
};
#endif /* !HAVE_MMALOC_H */

#ifdef CCXX_NAMESPACES
namespace damldb {
#endif

  template <class T> class MMallocAllocator
  {
  public:
    typedef T                 value_type;
    typedef value_type*       pointer;
    typedef const value_type* const_pointer;
    typedef value_type&       reference;
    typedef const value_type& const_reference;
    typedef size_t            size_type;
    typedef ptrdiff_t         difference_type;
                           
    template <class U> 
    struct rebind { typedef MMallocAllocator<U> other; };

    MMallocAllocator()
      : _file_name(""),
	_file_des(-1),
	_mmap_des(0),
	_base_address(0),
	_owns_mmap(true)
    {
      _mmap_des = ::mmalloc_attach(_file_des, NULL);
    }

    MMallocAllocator(const char* file_name,
		      const int flags = O_RDWR | O_CREAT,
		      const mode_t mode = 0777)
      : _file_name(file_name),
	_file_des(-1),
	_mmap_des(0),
	_base_address(0),
	_owns_mmap(true)
    {
      _file_des = ::open(_file_name.c_str(), flags, mode);
      if (_file_des < 0)
	{	THROW(runtime_error("constructor: open failed")); }

      _mmap_des = ::mmalloc_attach(_file_des, NULL);
    }

    MMallocAllocator(const MMallocAllocator& mm)
      : _file_name(mm._file_name),
	_file_des(mm._file_des),
	_mmap_des(mm._mmap_des),
	_base_address(mm._base_address),
	_owns_mmap(mm._owns_mmap)
    { mm._owns_mmap = false; }

    template <class U> 
    MMallocAllocator(const MMallocAllocator<U>& mm)
      : _file_name(mm._file_name),
	_file_des(mm._file_des),
	_mmap_des(mm._mmap_des),
	_base_address(mm._base_address),
	_owns_mmap(mm._owns_mmap)
    { mm._owns_mmap = false; }

    ~MMallocAllocator()
    {
      if (_owns_mmap)
	{
	  ::mmalloc_detach(_mmap_des);
	  _mmap_des = NULL;
	}
    }

    pointer address(reference x) const { return &x; }
    const_pointer address(const_reference x) const { 
      return x;
    }

    pointer allocate(size_type n, const_pointer = 0) {
      void* p = ::mmalloc(_mmap_des, n * sizeof(T));
      if (!p)
	THROW(bad_alloc());
      _base_address = static_cast<pointer>(p);
      return static_cast<pointer>(p);
    }

    pointer reallocate(size_type n, const_pointer = 0) {
      void* p = ::mrealloc(_mmap_des, _base_address, n * sizeof(T));
      if (!p)
	THROW(bad_alloc());
      _base_address = static_cast<pointer>(p);
      return static_cast<pointer>(p);
    }

    void deallocate(pointer p, size_type) { ::mfree(_mmap_des, p); }

    size_type max_size() const { 
      return static_cast<size_type>(-1) / sizeof(T);
    }

    void construct(pointer p, const value_type& x) { 
      new(p) value_type(x); 
    }
    void destroy(pointer p) { p->~value_type(); }

    bool get_owns_mmap() const
    { return _owns_mmap; }

    void set_owns_mmap(const bool owns = true)
    { _owns_mmap = owns; }

  private:
    void operator=(const MMallocAllocator&);

  private:
    string	_file_name;
    int		_file_des;
    void*	_mmap_des;
    pointer	_base_address;
    mutable bool  _owns_mmap;
  };

  template<> class MMallocAllocator<void>
  {
    typedef void        value_type;
    typedef void*       pointer;
    typedef const void* const_pointer;

    template <class U> 
    struct rebind { typedef MMallocAllocator<U> other; };
  };


  template <class T>
  inline bool operator==(const MMallocAllocator<T>&, 
			 const MMallocAllocator<T>&) {
    return true;
  }

  template <class T>
  inline bool operator!=(const MMallocAllocator<T>&, 
			 const MMallocAllocator<T>&) {
    return false;
  }

#ifdef CCXX_NAMESPACES
} // namespace damldb
#endif

#endif /* _DAMLDB_MMALOC_ALLOCATOR_HH */

// Emacs
// Local Variables:
// mode: c++
// c-basic-offset: 2
// End:
