
#ifndef _DAMLDB_STATEMENT_ITERATOR_HH
#define _DAMLDB_STATEMENT_ITERATOR_HH

#if defined (sun)
#  pragma ident "@(#)$Id: StatementIterator.hh,v 1.2 2002/04/09 02:04:46 neves Exp $"
#endif

#include "damldb/IDamlDB.hh"
#include "damldb/Portability.hh"
#include "damldb/Statement.hh"

#ifdef CCXX_NAMESPACES
namespace damldb {
#endif

  /** StatementIteratorBase is the base class for all Statement record iterators.
  */
  struct StatementIteratorBase
  {
    typedef size_t               size_type;
    typedef ptrdiff_t            difference_type;
    typedef forward_iterator_tag iterator_category;

    // Pointer to a IDamlDB object
    IDamlDB* pDB;

    // Pointer to the current Statement record
    Statement* pStatement;

    /** StatementIteratorBase Constructor
	Construct a StatementIteratorBase object.
	@param db pointer to the IDamlDB object
	@param pStmt pointer to the first Statement record
     */
    StatementIteratorBase(IDamlDB *db, Statement* pStmt) : pDB(db), pStatement(pStmt) {}

    /* Advance pStatement to point to the next record. */
    void _advance()
    {
      if (pStatement != 0)
	{
	  pStatement = (pStatement->objectNext != NULL_STATEMENT_ID) ?
	    const_cast<Statement*>(&pDB->get_statement(pStatement->objectNext)) : 0;
	}
    }

    /** Compare two StatementIteratorBase objects for equality
	@param rhs the right hand StatementIteratorBase
	@return true, if the iterators are equal. Otherwise, false.
     */
    bool operator==(const StatementIteratorBase& rhs) const {
      return pStatement == rhs.pStatement;
    }

    /** Compare two StatementIteratorBase objects for inequality
	@param rhs the right hand StatementIteratorBase
	@return true, if the iterators are not equal. Otherwise, false.
     */
    bool operator!=(const StatementIteratorBase& rhs) const {
      return pStatement != rhs.pStatement;
    }
  };

  /** StatementIterator is an iterator over a subset of RDF
      statements.  The subset is all statements whose predicate is an
      rdf \it{Type} and the object must be an rdfs \it{Class}.
  */
  struct StatementIterator : public StatementIteratorBase
  {
    typedef StatementIterator		iterator;
    typedef const StatementIterator	const_iterator;

    typedef Statement*	   pointer;
    typedef Statement&	   reference;

    /** StatementIterator Constructor
	Construct a StatementIterator object.
	@param db pointer to the IDamlDB object
	@param pStmt pointer to the first Statement record
     */
    StatementIterator(IDamlDB* db, pointer pStmt) : StatementIteratorBase(db, pStmt) {}

    /** IDamlDB/Statement Constructor
	Construct a StatementIterator object that represents the iteration end.
     */
    StatementIterator() : StatementIteratorBase(0, 0) {}

    /** Copy Constructor
	Construct a copy of a StatementIterator object.
	@param iter pointer to the iterator to copy.
     */
    StatementIterator(const iterator& iter) : StatementIteratorBase(iter.pDB, iter.pStatement) {}

    /** Dereference operator
	@precondition *this' != StatementIterator'()
	@returns a reference to the current Statement
     */
    reference operator*() const { return *pStatement; }

    /** Arrow operator
	@precondition *this' != StatementIterator'()
	@returns a pointer to the current Statement
     */
    pointer operator->() const { return &(operator*()); }

    /** Prefix increment operator
     */
    iterator& operator++()
    {
      _advance();
      return *this;
    }

    /**  Postfix increment operator
     */
    iterator operator++(int)
    {
      iterator __tmp = *this;
      _advance();
      return __tmp;
    }
  };

  /**  Returns the distance between two iterators
       @param iter a StatementIteratorBase object
       @return the distance
   */
  inline ptrdiff_t* distance_type(const StatementIteratorBase&) {
    return 0;
  }

  /**  Returns a StatementIteratorBase's iterator_category
       @param iter a StatementIteratorBase object
       @return a iterator_tag object
   */
  inline forward_iterator_tag iterator_category(const StatementIteratorBase&) {
    return forward_iterator_tag();
  }

  /**  Returns the \emph{value_type} of the current iterator record
       @param iter a StatementIterator object
       @return a pointer to a Statement record
   */
  inline Statement* value_type(const StatementIterator&) {
    return 0;
  }

#ifdef CCXX_NAMESPACES
} // namespace damldb
#endif

#endif /* !_DAMLDB_STATEMENT_ITERATOR_HH */

// Emacs
// Local Variables:
// mode: c++
// c-basic-offset: 2
// End:

