#ifndef _DAMLDB_MMAP_ALLOCATOR_IMPL_HH
#define _DAMLDB_MMAP_ALLOCATOR_IMPL_HH

#if defined (sun)
#  pragma ident "@(#)$Id: MMapAllocatorImpl.hh,v 1.2 2002/04/11 22:33:21 neves Exp $"
#endif

#include "damldb/DBHeader.hh"
#include "damldb/MMapAllocator.hh"
#include "damldb/OS.hh"
#include "damldb/Portability.hh"
#include "damldb/damldb_iodb.hh"

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <memory>
#include <string>
#include <stdexcept>
#include <typeinfo>

#include <string.h>
#include <errno.h>

#include <sys/mman.h>
#if defined(HAVE_SYS_STAT_H)
#   include <sys/stat.h>
#endif

#ifdef CCXX_NAMESPACES
namespace damldb {
#endif

  template <class T> 
  class MMapAllocatorImpl : public MMapAllocator<T>
  {
  public:
    typedef T                 value_type;
    typedef value_type*       pointer;
    typedef const value_type* const_pointer;
    typedef value_type&       reference;
    typedef const value_type& const_reference;
    typedef size_t            size_type;
    typedef ptrdiff_t         difference_type;

    typedef int               fd_t;
    typedef char*	      byte_pointer;
    typedef const char*	      const_byte_pointer;

    template <class U> 
    struct rebind { typedef MMapAllocatorImpl<U> other; };

    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl()
      : MMapAllocator<T>(),
	_file_des(-1),
	_open_mode(iodb::in | iodb::out | iodb::nocreate),
	_base_address(0),
	_file_size(0),
	_class_name(typeid(T).name()),
	_dir_name("."),
	_file_name(typeid(T).name()),
	_close_file(true)
    {
      _file_des  = maybe_create_file();
      _file_size = get_file_size(_file_des);

      if (_file_size == 0) {
	THROW(runtime_error("Invalid File database file specfied"));
      }

      (void) create_mapped_view(&_base_address);
    }
    
    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl(const char* dir_name,
			 const char* file_name,
			 int open_mode = iodb::in | iodb::out | iodb::nocreate)
      : MMapAllocator<T>(),
	_file_des(-1),
	_open_mode(open_mode),
	_base_address(0),
	_file_size(0),
	_class_name(typeid(T).name()),
	_dir_name(dir_name),
	_file_name(file_name),
	_close_file(true)
    {
      _file_des  = maybe_create_file();
      _file_size = get_file_size(_file_des);

      if (_file_size == 0) {
	THROW(runtime_error("Invalid File database file specfied"));
      }

      (void) create_mapped_view(&_base_address);
    }

    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl(const MMapAllocatorImpl& w)
      : MMapAllocator<T>(w),
	_file_des(w._file_des),
	_open_mode(w._open_mode),
	_base_address(w._base_address),
	_file_size(w._file_size),
	_class_name(w._class_name),
	_dir_name(w._dir_name),
	_file_name(w._file_name),
	_close_file(w._close_file)
    {}

    //////////////////////////////////////////////////////////////////////

    template <class U> 
    MMapAllocatorImpl(const MMapAllocatorImpl<U>& w)
      : MMapAllocator<U>(w),
	_file_des(w._file_des),
	_open_mode(w._open_mode),
	_base_address(w._base_address),
	_file_size(w._file_size),
	_class_name(w._class_name),
	_dir_name(w._dir_name),
	_file_name(w._file_name),
	_close_file(w._close_file)
    {}

    //////////////////////////////////////////////////////////////////////

    virtual ~MMapAllocatorImpl()
    {
      if (_close_file) {
	const caddr_t start = reinterpret_cast<caddr_t>(_base_address);
	const size_type length = _file_size;
	int retval = 0;

	if (start != NULL)
	  {
	    retval = munmap(start, length);
	    if (retval < 0)
	      { THROW(runtime_error(::strerror(errno))); }
	  }

	if (_file_des != -1)
	  {
	    retval = close(_file_des);
	    if (retval < 0)
	      { THROW(runtime_error(::strerror(errno))); }
	    _file_des = -1;
	  }
      }
    }

    //////////////////////////////////////////////////////////////////////

    pointer allocate(size_type n, const_pointer = 0)
    {
      size_t nbytes = n * sizeof(T) + sizeof(DBHeader);

      if (nbytes > _file_size) {
	reallocate(n);
      }

      return reinterpret_cast<pointer>(((byte_pointer)_base_address) + sizeof(DBHeader));
    }

    //////////////////////////////////////////////////////////////////////

    pointer reallocate(size_type n, const_pointer = 0)
    {
      size_t nbytes = n * sizeof(T) + sizeof(DBHeader);

      if (nbytes <= _file_size) {
	return reinterpret_cast<pointer>(((byte_pointer)_base_address) + sizeof(DBHeader));
      }
  
      if (_base_address != NULL) {
	const caddr_t start = reinterpret_cast<caddr_t>(_base_address);
	const size_type length = _file_size;

	if (munmap(start, length) < 0)
	  {
	    THROW(runtime_error(::strerror(errno)));
	  }

	_base_address = 0;
      }

      const size_type new_nbytes = nbytes - _file_size;
      char* buf = new char[new_nbytes];
      try
	{
	  memset(buf, 0, new_nbytes);
	  append_buffer_to_file(_file_des, buf, new_nbytes);
	}
      catch(const exception& e)
	{
	  delete [] buf;
	  close(_file_des);
	  THROW(e);
	}

      delete [] buf;

      if (close(_file_des) < 0)
	{
	  THROW(runtime_error(::strerror(errno)));
	}

      _open_mode   = _open_mode | iodb::nocreate;
      _file_des    = maybe_create_file();
      _file_size   = get_file_size(_file_des);
      (void) create_mapped_view(&_base_address);

      return reinterpret_cast<pointer>(((byte_pointer)_base_address) + sizeof(DBHeader));
    }

    //////////////////////////////////////////////////////////////////////

    void deallocate(pointer p, size_type) {}

    //////////////////////////////////////////////////////////////////////

    pointer end_of_storage() const
    { return reinterpret_cast<pointer>(((byte_pointer)_base_address) + _file_size); }

    //////////////////////////////////////////////////////////////////////

    size_type get_file_size(fd_t file_handle) const
    {
      struct stat stat;
      int retVal = 0;

      if (fstat(file_handle, &stat) < 0) {
	THROW(runtime_error(::strerror(errno)));
      }

      retVal = stat.st_size;

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    size_type get_record_count() const {
      DBHeader* header = reinterpret_cast<DBHeader*>(_base_address);
      return header->size;
    }

    //////////////////////////////////////////////////////////////////////

    void set_record_count(size_type size) {
      DBHeader* header = reinterpret_cast<DBHeader*>(_base_address);
      header->size = size;
    }

    //////////////////////////////////////////////////////////////////////

    void set_close_file_on_destroy(bool close_file = true) const {
      _close_file = close_file;
    }

  private:

    //////////////////////////////////////////////////////////////////////

    int desired_access(int open_mode) const
    {
      int retVal = 0;

      if (open_mode & iodb::in)
	{
	  retVal |= O_RDONLY;
	}

      if (open_mode & iodb::out)
	{
	  retVal |= O_WRONLY;
	}

      if ((open_mode & iodb::in) && (open_mode & iodb::out))
	{
	  retVal = O_RDWR;
	}

      if (!(open_mode & iodb::nocreate))
	{
	  retVal |= O_CREAT | O_EXCL;
	}

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    int desired_map_access(int open_mode) const
    {
      int retVal = PROT_READ;

      if (open_mode & iodb::in && open_mode & iodb::out) {
	retVal = PROT_READ | PROT_WRITE;
      }

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    fd_t create_mapped_view(pointer* base_address)
    {
      const caddr_t start	= 0;
      const size_type length	= _file_size;
      const int prot		= desired_map_access(_open_mode);
      const int flags		= MAP_SHARED;
      const size_type offset	= 0;

      void* mapped_base_addr =
	mmap(start, length, prot, flags, _file_des, offset);

      if (mapped_base_addr == MAP_FAILED) {
	THROW(runtime_error(::strerror(errno)));
      }

      *base_address =
	static_cast<pointer>(mapped_base_addr);

      return _file_des;
    }

    //////////////////////////////////////////////////////////////////////

    fd_t maybe_create_file() const
    {
      fd_t retVal = -1;
      auto_ptr<const string> file_name(get_full_path_name());

      const char* fileName = file_name->c_str();
      const int      flags = desired_access(_open_mode);
      const mode_t   mode  = 0777;

      // create file if it doesn't exist already
      retVal = open(fileName, flags, mode);
      if (retVal < 0) {
	char* msg = NULL;
	if (_open_mode & iodb::nocreate)
	  {
	    msg = "Database does not exist";
	  }
	else
	  {
	    msg = "Database already exists";
	  }

	THROW(runtime_error(msg));
      }

      try
	{
	  const size_type file_size = get_file_size(retVal);
	  if (file_size == 0 && !(_open_mode & iodb::nocreate)) {
	    write_file_header(retVal, 0);
	  }
	} 
      catch(const runtime_error& e)
	{
	  ::close(retVal);
	  THROW(runtime_error(e.what()));
	}

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    const string& get_class_name() const
    {
      return _class_name;
    }

    //////////////////////////////////////////////////////////////////////

    const string* get_full_path_name() const
    {
      const string* path_name = new string( _dir_name.empty() ?
						      _file_name :
						      _dir_name + "/" + _file_name );
      return path_name;
    }

    //////////////////////////////////////////////////////////////////////

    void append_buffer_to_file(fd_t filedes, const char* buf, const size_type nbytes) const
    {
      off_t offset = 0;
      int   whence = SEEK_END;

      if (lseek(filedes, offset, whence) < 0) {
	THROW(runtime_error(::strerror(errno)));
      }

      write_buffer_to_file(filedes, buf, nbytes);
    }

    //////////////////////////////////////////////////////////////////////

    void write_buffer_to_file(fd_t filedes, const char* buf, const size_type nbytes) const
    {
      char* cp = const_cast<char*>(buf);
      size_t toWrite = nbytes;

      while (toWrite > 0) {
	int written;
	if ((written = write(filedes, cp, toWrite)) < 0) {
	  THROW(runtime_error(::strerror(errno)));
	}
	toWrite -= written;
	cp += written;
      }
    }

    //////////////////////////////////////////////////////////////////////

    void write_file_header(fd_t filedes, const size_type n) const
    {
      DBHeader header;

      memcpy(&header.magic, "@damldb@", sizeof(header.magic));
      header.major_version = 0;
      header.minor_version = 0;
      header.size          = n;

      append_buffer_to_file(filedes, reinterpret_cast<char*>(&header), sizeof(header));
    }

  private:

    void operator=(const MMapAllocatorImpl&);

  private:
    fd_t	_file_des;
    int		_open_mode;
    pointer	_base_address;
    size_type	_file_size;
    string	_class_name;
    string	_dir_name;
    string	_file_name;
    mutable bool _close_file;
  };

  template<> class MMapAllocatorImpl<void>
  {
    typedef void        value_type;
    typedef void*       pointer;
    typedef const void* const_pointer;

    template <class U> 
    struct rebind { typedef MMapAllocatorImpl<U> other; };
  };


  template <class T>
  inline bool operator==(const MMapAllocatorImpl<T>&, 
			 const MMapAllocatorImpl<T>&) {
    return false;
  }

  template <class T>
  inline bool operator!=(const MMapAllocatorImpl<T>&, 
			 const MMapAllocatorImpl<T>&) {
    return true;
  }

#ifdef CCXX_NAMESPACES
} // namespace damldb
#endif

#endif /* !_DAMLDB_MMAP_ALLOCATOR_IMPL_HH */

// Emacs
// Local Variables:
// mode: c++
// c-basic-offset: 2
// End:
