#ifndef _DAMLDB_MMAP_ALLOCATOR_IMPL_HH
#define _DAMLDB_MMAP_ALLOCATOR_IMPL_HH

#include "damldb/DBHeader.hh"
#include "damldb/MMapAllocator.hh"
#include "damldb/Portability.hh"
#include "damldb/damldb_iodb.hh"
#include "damldb/support.h"

#include <iostream>
#include <memory>
#include <stdexcept>
#include <typeinfo>
#include <windows.h>

#ifdef CCXX_NAMESPACES
namespace damldb {
USING(std);
#endif

  template <class T> 
  class MMapAllocatorImpl : public MMapAllocator<T>
  {
  public:
    typedef T                 value_type;
    typedef value_type*       pointer;
    typedef const value_type* const_pointer;
    typedef value_type&       reference;
    typedef const value_type& const_reference;
    typedef size_t            size_type;
    typedef ptrdiff_t         difference_type;

    template <class U> 
    struct rebind { typedef MMapAllocatorImpl<U> other; };

    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl()
      : MMapAllocator<T>(),
	_file_handle(INVALID_HANDLE_VALUE),
	_map_handle(0),
	_open_mode(iodb::in | iodb::out | iodb::nocreate),
	_base_address(0),
	_file_size(0),
	_class_name(typeid(T).name()),
	_dir_name("."),
	_file_name(typeid(T).name()),
	_close_file(true)
    {
      _file_handle = maybe_create_file();
      _file_size   = get_file_size(_file_handle);

      if (_file_size == 0) {
	THROW(runtime_error("Invalid File database file specfied"));
      }

      _map_handle  = create_mapped_view(&_base_address);
    }
    
    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl(const char* dir_name,
			 const char* file_name,
			 int open_mode = iodb::in | iodb::out | iodb::nocreate)
      : MMapAllocator<T>(),
	_file_handle(0),
	_map_handle(0),
	_open_mode(open_mode),
	_base_address(0),
	_file_size(0),
	_class_name(typeid(T).name()),
	_dir_name(dir_name),
	_file_name(file_name),
	_close_file(true)
    {
      _file_handle = maybe_create_file();
      _file_size   = get_file_size(_file_handle);

      if (_file_size == 0) {
	THROW(runtime_error("Invalid File database file specfied"));
      }

      _map_handle  = create_mapped_view(&_base_address);
    }

    //////////////////////////////////////////////////////////////////////

    MMapAllocatorImpl(const MMapAllocatorImpl& w)
      : MMapAllocator<T>(w),
	_file_handle(w._file_handle),
	_map_handle(w._map_handle),
	_open_mode(w._open_mode),
	_base_address(w._base_address),
	_file_size(w._file_size),
	_class_name(w._class_name),
	_dir_name(w._dir_name),
	_file_name(w._file_name),
	_close_file(w._close_file)
    {}

    //////////////////////////////////////////////////////////////////////

#if !defined(_WIN32)	// MS Visual C++ does not like this
    template <class U> 
    MMapAllocatorImpl(const MMapAllocatorImpl<U>& w)
      : MMapAllocator<U>(w),
	_file_handle(w._file_handle),
	_map_handle(w._map_handle),
	_open_mode(w._open_mode),
	_base_address(w._base_address),
	_file_size(w._file_size),
	_class_name(w._class_name),
	_dir_name(w._dir_name),
	_file_name(w._file_name),
	_close_file(w._close_file)
    {}
#endif

    //////////////////////////////////////////////////////////////////////

    virtual ~MMapAllocatorImpl()
    {
      if (_close_file) {
	if (_map_handle != NULL) {
	  CloseHandle(_map_handle);
	  _map_handle = NULL;
	}

	if (_file_handle != INVALID_HANDLE_VALUE) {
	  CloseHandle(_file_handle);
	  _file_handle = NULL;
	}
      }
    }

    //////////////////////////////////////////////////////////////////////

    pointer allocate(size_type n, const_pointer = 0)
    {
      size_t nbytes = n * sizeof(T) + sizeof(DBHeader);

      if (nbytes > _file_size) {
	reallocate(n);
      }

      return reinterpret_cast<pointer>(((unsigned int)_base_address) + sizeof(DBHeader));
    }

    //////////////////////////////////////////////////////////////////////

    pointer reallocate(size_type n, const_pointer = 0)
    {
      size_t nbytes = n * sizeof(T) + sizeof(DBHeader);

      if (nbytes <= _file_size) {
	return reinterpret_cast<pointer>(((unsigned int)_base_address) + sizeof(DBHeader));
      }
  
      if (_base_address != NULL) {
	if (UnmapViewOfFile(_base_address) == 0) {
	}

	if (CloseHandle(_map_handle) == 0) {
	  THROW((runtime_error("reallocate: CloseHandle failed")));
	}

	_base_address = 0;;
      }

      const size_type new_nbytes = nbytes - _file_size;
      char* buf = new char[new_nbytes];
      try
	{
	  memset(buf, 0, new_nbytes);
	  append_buffer_to_file(_file_handle, buf, new_nbytes);
	}
      catch(const exception& e)
	{
	  delete [] buf;
	  CloseHandle(_file_handle);
	  THROW(e);
	}

      delete [] buf;
      CloseHandle(_file_handle);

      _open_mode   = _open_mode | iodb::nocreate;
      _file_handle = maybe_create_file();
      _file_size   = get_file_size(_file_handle);
      _map_handle  = create_mapped_view(&_base_address);

      return reinterpret_cast<pointer>(((unsigned int)_base_address) + sizeof(DBHeader));
    }

    //////////////////////////////////////////////////////////////////////

    void deallocate(pointer p, size_type) {}

    //////////////////////////////////////////////////////////////////////

    pointer end_of_storage() const
    { return reinterpret_cast<pointer>(((unsigned int)_base_address) + _file_size); }

    //////////////////////////////////////////////////////////////////////

    size_type get_file_size(HANDLE file_handle) const
    {
      DWORD retVal = 0;

      if ((retVal = GetFileSize(file_handle, NULL)) == INVALID_FILE_SIZE) {
	THROW(runtime_error(win32_strerror()));
      }

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    size_type get_record_count() const {
      DBHeader* header = reinterpret_cast<DBHeader*>(_base_address);
      return header->size;
    }

    //////////////////////////////////////////////////////////////////////

    void set_record_count(size_type size) {
      DBHeader* header = reinterpret_cast<DBHeader*>(_base_address);
      header->size = size;
    }

    //////////////////////////////////////////////////////////////////////

    void set_close_file_on_destroy(bool close_file = true) const {
      _close_file = close_file;
    }

  private:

    //////////////////////////////////////////////////////////////////////

    DWORD creation_distribution(int open_mode) const
    {
      return (open_mode & iodb::nocreate) ? OPEN_EXISTING : CREATE_NEW;
    }

    //////////////////////////////////////////////////////////////////////

    DWORD desired_access(int open_mode) const
    {
      DWORD retVal = 0;

      if (open_mode & iodb::in) {
	retVal |= GENERIC_READ;
      }

      if (open_mode & iodb::out) {
	retVal |= GENERIC_WRITE;
      }
      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    DWORD desired_map_access(int open_mode) const
    {
      DWORD retVal = PAGE_READONLY;

      if (open_mode & iodb::in && open_mode & iodb::out) {
	retVal = PAGE_READWRITE;
      }

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    DWORD desired_map_view_access(int open_mode) const
    {
      DWORD retVal = FILE_MAP_READ;

      if (open_mode & iodb::in && open_mode & iodb::out) {
	retVal = FILE_MAP_ALL_ACCESS;
      }

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    HANDLE create_mapped_view(pointer* base_address)
    {
      HANDLE map_handle = NULL;

      map_handle = CreateFileMapping(_file_handle,
				     NULL,
				     desired_map_access(_open_mode),
				     0,
				     0,
				     NULL);

      if (map_handle == NULL) {
	THROW(runtime_error(win32_strerror()));
      }

      void* map_base_address = MapViewOfFile(map_handle,
					     desired_map_view_access(_open_mode),
					     0,
					     0,
					     (SIZE_T)0);
      if (map_base_address == NULL) {
	THROW(runtime_error(win32_strerror()));
      }
  
      *base_address = static_cast<pointer>(map_base_address);
      return map_handle;
    }

    //////////////////////////////////////////////////////////////////////

    HANDLE maybe_create_file() const
    {
      HANDLE retVal = INVALID_HANDLE_VALUE;
      auto_ptr<const string> file_name(get_full_path_name());

      LPCTSTR lpFileName = file_name->c_str();
      DWORD dwDesiredAccess = desired_access(_open_mode);
      DWORD dwShareMode = 0;
      LPSECURITY_ATTRIBUTES lpSecurityAttributes = NULL;
      DWORD dwCreationDistribution = creation_distribution(_open_mode);
      DWORD dwFlagsAndAttributes = FILE_FLAG_RANDOM_ACCESS;
      HANDLE hTemplateFile = NULL;

      // create file if it doesn't exist already
      retVal = CreateFile(lpFileName,
			  dwDesiredAccess,
			  dwShareMode,
			  lpSecurityAttributes,
			  dwCreationDistribution,
			  dwFlagsAndAttributes,
			  hTemplateFile);

      if (retVal == INVALID_HANDLE_VALUE)
	{
	  char* msg = NULL;
	  if (_open_mode & iodb::nocreate)
	    {
	      msg = "Database does not exist";
	    }
	  else
	    {
	      msg = "Database already exists";
	    }

	  THROW(runtime_error(msg));
	}

      try
	{
	  const size_type file_size = get_file_size(retVal);
	  if (file_size == 0 && !(_open_mode & iodb::nocreate)) {
	    write_file_header(retVal, 0);
	  }
	}
      catch(const runtime_error& e)
	{
	  CloseHandle(retVal);
	  THROW(runtime_error("Unable to write damldb metadata"));
	}

      return retVal;
    }

    //////////////////////////////////////////////////////////////////////

    const string& get_class_name() const
    {
      return _class_name;
    }

    //////////////////////////////////////////////////////////////////////

    const string* get_full_path_name() const
    {
      const string* path_name = new string( _dir_name.empty() ?
						      _file_name :
						      _dir_name + "/" + _file_name );
      return path_name;
    }

    //////////////////////////////////////////////////////////////////////

    void append_buffer_to_file(HANDLE hFile, const char* buf, const size_type nbytes) const
    {
      SetFilePointer(hFile, 0, NULL, FILE_END);
      write_buffer_to_file(hFile, buf, nbytes);
    }

    //////////////////////////////////////////////////////////////////////

    void write_buffer_to_file(HANDLE hFile, const char* buf, const size_type nbytes) const
    {
      char* cp = const_cast<char*>(buf);
      size_t toWrite = nbytes;

      while (toWrite > 0) {
	DWORD written;

	if (WriteFile(hFile, cp, toWrite, & written, NULL) == 0) {
	  THROW(runtime_error(win32_strerror()));
	}
	toWrite -= written;
	cp += written;
      }
    }

    //////////////////////////////////////////////////////////////////////

    void write_file_header(HANDLE hFile, const size_type n) const
    {
      DBHeader header;

      ::memcpy(&header.magic, "@damldb@", sizeof(header.magic));
      header.major_version = 0;
      header.minor_version = 0;
      header.size          = n;

      append_buffer_to_file(hFile, reinterpret_cast<char*>(&header), sizeof(header));
    }

  private:
    void operator=(const MMapAllocatorImpl&);

  private:
    HANDLE	_file_handle;
    HANDLE	_map_handle;
    int           _open_mode;
    pointer       _base_address;
    size_type	_file_size;
    string	_class_name;
    string	_dir_name;
    string	_file_name;
    mutable bool	_close_file;
  };

  template<> class MMapAllocatorImpl<void>
  {
    typedef void        value_type;
    typedef void*       pointer;
    typedef const void* const_pointer;

    template <class U> 
    struct rebind { typedef MMapAllocatorImpl<U> other; };
  };


  template <class T>
  inline bool operator==(const MMapAllocatorImpl<T>&, 
			 const MMapAllocatorImpl<T>&) {
    return false;
  }

  template <class T>
  inline bool operator!=(const MMapAllocatorImpl<T>&, 
			 const MMapAllocatorImpl<T>&) {
    return true;
  }

#ifdef CCXX_NAMESPACES
} // namespace damldb
#endif

#endif /* !_DAMLDB_MMAP_ALLOCATOR_IMPL_HH */

// Emacs
// Local Variables:
// mode: c++
// c-basic-offset: 2
// End:
